@@ -190,16 +190,13 @@ convert_cmyk_to_rgb (struct jpeg_decompress_struct *cinfo,
 			m = p[1];
 			y = p[2];
 			k = p[3];
-			if (cinfo->saw_Adobe_marker) {
-				p[0] = k*c / 255;
-				p[1] = k*m / 255;
-				p[2] = k*y / 255;
-			}
-			else {
-				p[0] = (255 - k)*(255 - c) / 255;
-				p[1] = (255 - k)*(255 - m) / 255;
-				p[2] = (255 - k)*(255 - y) / 255;
-			}
+
+			/* We now assume that all CMYK JPEG files
+			 * use inverted CMYK, as Photoshop does
+			 * See https://bugzilla.gnome.org/show_bug.cgi?id=618096 */
+			p[0] = k*c / 255;
+			p[1] = k*m / 255;
+			p[2] = k*y / 255;
 			p[3] = 255;
 			p += 4;
 		}
@@ -470,18 +467,23 @@ jpeg_parse_exif_app1 (JpegExifContext *context, jpeg_saved_marker_ptr marker)
 		guint tag   = de_get16(&marker->data[i + 0], endian);
 		guint type  = de_get16(&marker->data[i + 2], endian);
 		guint count = de_get32(&marker->data[i + 4], endian);
-		/* values of types small enough to fit are stored directly in the (first) bytes of the Value Offset field */
-		guint short_value = de_get16(&marker->data[i + 8], endian);
 
 		/* orientation tag? */
 		if (tag == 0x112){
 
-			/* The orientation field should consist of a single 2-byte integer */
-			if (type != 0x3 || count != 1)
-				continue;
+			/* The orientation field should consist of a single 2-byte integer,
+			 * but might be a signed long.
+			 * Values of types smaller than 4 bytes are stored directly in the
+			 * Value Offset field */
+			if (type == 0x3 && count == 1) {
+				guint short_value = de_get16(&marker->data[i + 8], endian);
 
-			/* get the orientation value */
-			context->orientation = short_value <= 8 ? short_value : 0;
+				context->orientation = short_value <= 8 ? short_value : 0;
+			} else if (type == 0x9 && count == 1) {
+				guint long_value = de_get32(&marker->data[i + 8], endian);
+
+				context->orientation = long_value <= 8 ? long_value : 0;
+			}
 		}
 		/* move the pointer to the next 12-byte tag field. */
 		i = i + 12;
@@ -519,6 +521,7 @@ gdk_pixbuf__jpeg_image_load (FILE *f, GError **error)
 {
 	gint   i;
 	char   otag_str[5];
+	char  *density_str;
 	GdkPixbuf * volatile pixbuf = NULL;
 	guchar *dptr;
 	guchar *lines[4]; /* Used to expand rows, via rec_outbuf_height, 
@@ -600,6 +603,27 @@ gdk_pixbuf__jpeg_image_load (FILE *f, GError **error)
 		goto out; 
 	}
 
+	switch (cinfo.density_unit) {
+	case 1:
+		/* Dots per inch (no conversion required) */
+		density_str = g_strdup_printf ("%d", cinfo.X_density);
+		gdk_pixbuf_set_option (pixbuf, "x-dpi", density_str);
+		g_free (density_str);
+		density_str = g_strdup_printf ("%d", cinfo.Y_density);
+		gdk_pixbuf_set_option (pixbuf, "y-dpi", density_str);
+		g_free (density_str);
+		break;
+	case 2:
+		/* Dots per cm - convert into dpi */
+		density_str = g_strdup_printf ("%d", DPCM_TO_DPI (cinfo.X_density));
+		gdk_pixbuf_set_option (pixbuf, "x-dpi", density_str);
+		g_free (density_str);
+		density_str = g_strdup_printf ("%d", DPCM_TO_DPI (cinfo.Y_density));
+		gdk_pixbuf_set_option (pixbuf, "y-dpi", density_str);
+		g_free (density_str);
+		break;
+	}
+
 	/* if orientation tag was found */
 	if (exif_context.orientation != 0) {
 		g_snprintf (otag_str, sizeof (otag_str), "%d", exif_context.orientation);
@@ -901,6 +925,7 @@ gdk_pixbuf__jpeg_image_load_increment (gpointer data,
 	gint             width, height;
 	char             otag_str[5];
 	gchar 		*icc_profile_base64;
+	char            *density_str;
 	JpegExifContext  exif_context = { 0, };
 	gboolean	 retval;
 
@@ -1038,6 +1063,27 @@ gdk_pixbuf__jpeg_image_load_increment (gpointer data,
                                 retval = FALSE;
 				goto out;
 			}
+
+			switch (cinfo->density_unit) {
+			case 1:
+				/* Dots per inch (no conversion required) */
+				density_str = g_strdup_printf ("%d", cinfo->X_density);
+				gdk_pixbuf_set_option (context->pixbuf, "x-dpi", density_str);
+				g_free (density_str);
+				density_str = g_strdup_printf ("%d", cinfo->Y_density);
+				gdk_pixbuf_set_option (context->pixbuf, "y-dpi", density_str);
+				g_free (density_str);
+				break;
+			case 2:
+				/* Dots per cm - convert into dpi */
+				density_str = g_strdup_printf ("%d", DPCM_TO_DPI (cinfo->X_density));
+				gdk_pixbuf_set_option (context->pixbuf, "x-dpi", density_str);
+				g_free (density_str);
+				density_str = g_strdup_printf ("%d", DPCM_TO_DPI (cinfo->Y_density));
+				gdk_pixbuf_set_option (context->pixbuf, "y-dpi", density_str);
+				g_free (density_str);
+				break;
+			}
 		
 		        /* if orientation tag was found set an option to remember its value */
 			if (exif_context.orientation != 0) {
@@ -1226,6 +1272,8 @@ real_save_jpeg (GdkPixbuf          *pixbuf,
        int n_channels;
        struct error_handler_data jerr;
        ToFunctionDestinationManager to_callback_destmgr;
+       int x_density = 0;
+       int y_density = 0;
        gchar *icc_profile = NULL;
        gchar *data;
        gint retval = TRUE;
@@ -1265,6 +1313,48 @@ real_save_jpeg (GdkPixbuf          *pixbuf,
                                                     _("JPEG quality must be a value between 0 and 100; value '%d' is not allowed."),
                                                     quality);
 
+                                       retval = FALSE;
+                                       goto cleanup;
+                               }
+                       } else if (strcmp (*kiter, "x-dpi") == 0) {
+                               char *endptr = NULL;
+                               x_density = strtol (*viter, &endptr, 10);
+                               if (endptr == *viter)
+                                       x_density = -1;
+
+                               if (x_density <= 0 ||
+                                   x_density > 65535) {
+                                       /* This is a user-visible error;
+                                        * lets people skip the range-checking
+                                        * in their app.
+                                        */
+                                       g_set_error (error,
+                                                    GDK_PIXBUF_ERROR,
+                                                    GDK_PIXBUF_ERROR_BAD_OPTION,
+                                                    _("JPEG x-dpi must be a value between 1 and 65535; value '%s' is not allowed."),
+                                                    *viter);
+
+                                       retval = FALSE;
+                                       goto cleanup;
+                               }
+                       } else if (strcmp (*kiter, "y-dpi") == 0) {
+                               char *endptr = NULL;
+                               y_density = strtol (*viter, &endptr, 10);
+                               if (endptr == *viter)
+                                       y_density = -1;
+
+                               if (y_density <= 0 ||
+                                   y_density > 65535) {
+                                       /* This is a user-visible error;
+                                        * lets people skip the range-checking
+                                        * in their app.
+                                        */
+                                       g_set_error (error,
+                                                    GDK_PIXBUF_ERROR,
+                                                    GDK_PIXBUF_ERROR_BAD_OPTION,
+                                                    _("JPEG y-dpi must be a value between 1 and 65535; value '%s' is not allowed."),
+                                                    *viter);
+
                                        retval = FALSE;
                                        goto cleanup;
                                }
@@ -1355,6 +1445,13 @@ real_save_jpeg (GdkPixbuf          *pixbuf,
        jpeg_set_defaults (&cinfo);
        jpeg_set_quality (&cinfo, quality, TRUE);
 
+       /* set density information */
+       if (x_density > 0 && y_density > 0) {
+           cinfo.density_unit = 1; /* Dots per inch */
+           cinfo.X_density = x_density;
+           cinfo.Y_density = y_density;
+       }
+
        jpeg_start_compress (&cinfo, TRUE);
 
 	/* write ICC profile data */
@@ -1475,7 +1572,7 @@ MODULE_ENTRY (fill_info) (GdkPixbufFormat *info)
 
 	info->name = "jpeg";
 	info->signature = (GdkPixbufModulePattern *) signature;
-	info->description = N_("The JPEG image format");
+	info->description = NC_("image format", "JPEG");
 	info->mime_types = (gchar **) mime_types;
 	info->extensions = (gchar **) extensions;
 	info->flags = GDK_PIXBUF_FORMAT_WRITABLE | GDK_PIXBUF_FORMAT_THREADSAFE;
